use itertools::Itertools;
use nom::{
    branch::alt,
    bytes::complete::tag,
    bytes::complete::{tag_no_case, take_until},
    character::complete::newline,
    character::complete::space1,
    combinator::opt,
    error::VerboseError,
    IResult,
};

use crate::cpu::instructions::Instruction;
use crate::interpreter::parser::primitives::{
    parse_args, parse_empty_line, parse_full_data_source, parse_name,
    parse_usize_value,
};
use crate::interpreter::parser::CodeLine;

pub fn mov_instruction(
    line: &str,
) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("MOV")(line)?;
    let (out, _) = space1(out)?;
    let (out, args) = parse_args(2, parse_full_data_source)(out)?;
    let (op1, op2) = args.into_iter().next_tuple().unwrap();
    debug!("Parsed MOV instruction");
    Ok((out, Instruction::MOV { from: op1, to: op2 }))
}

pub fn ldr_instruction(
    line: &str,
) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("LDR")(line)?;
    let (out, _) = space1(out)?;
    let (out, dest) = parse_full_data_source(out)?;
    let (out, _) = space1(out)?;
    let (out, source) = parse_usize_value(out)?;
    debug!("Parsed LDR instruction");
    Ok((out, Instruction::LDR { source, dest }))
}

pub fn str_instruction(
    line: &str,
) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("STR")(line)?;
    let (out, _) = space1(out)?;
    let (out, source) = parse_full_data_source(out)?;
    let (out, _) = space1(out)?;
    let (out, dest) = parse_usize_value(out)?;
    debug!("Parsed STR instruction");
    Ok((out, Instruction::STR { source, dest }))
}

pub fn mark_inst(inp: &str) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("MARK")(inp)?;
    let (out, _) = space1(out)?;
    let (out, name) = parse_name(out)?;
    debug!("Parsed MARK {}", name.as_str());
    Ok((out, Instruction::MARK { name }))
}

pub fn call_inst(inp: &str) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("CALL")(inp)?;
    let (out, _) = space1(out)?;
    let (out, name) = parse_name(out)?;
    debug!("Parsed CALL {}", name.as_str());
    Ok((out, Instruction::CALL { name }))
}

pub fn jump_inst(inp: &str) -> IResult<&str, Instruction, VerboseError<&str>> {
    let (out, _) = tag_no_case("JUMP")(inp)?;
    let (out, _) = space1(out)?;
    let (out, mark) = parse_name(out)?;
    debug!("Parsed Jump to {}", mark.as_str());
    Ok((out, Instruction::JUMP { mark }))
}

pub fn parse_comment(
    inp: &str,
) -> IResult<&str, Option<CodeLine>, VerboseError<&str>> {
    let (out, _) = tag("#")(inp)?;
    let (out, _) = opt(space1)(out)?;
    let (out, comment) = take_until("\n")(out)?;
    let (out, _) = opt(newline)(out)?;
    Ok((out, Some(CodeLine::Comment(String::from(comment)))))
}

pub fn parse_instuction(
    inp: &str,
) -> IResult<&str, Option<CodeLine>, VerboseError<&str>> {
    let (out, line) = alt((
        mov_instruction,
        mark_inst,
        jump_inst,
        call_inst,
        str_instruction,
        ldr_instruction,
    ))(inp)?;
    Ok((out, Some(CodeLine::Code(line))))
}

pub fn parse_line(
    line: &str,
) -> IResult<&str, Option<CodeLine>, VerboseError<&str>> {
    alt((parse_comment, parse_instuction, parse_empty_line))(line)
}
