#[macro_use]
extern crate failure;
#[macro_use]
extern crate log;
#[macro_use]
extern crate nom;
#[cfg(test)]
extern crate quickcheck;
#[cfg(test)]
#[macro_use(quickcheck)]
extern crate quickcheck_macros;

use std::path::PathBuf;
use std::str::FromStr;

use fern::colors;
use log::LevelFilter;
use structopt::StructOpt;

use crate::cpu::model::CPU;
use crate::msl_result::MSLResult;

pub mod cpu;
pub mod ram;

pub mod interpreter;
pub mod msl_result;
pub mod repl;

#[cfg(test)]
mod tests {
    fn reverse<T: Clone>(xs: &[T]) -> Vec<T> {
        let mut rev = vec![];
        println!("test");
        for x in xs {
            rev.insert(0, x.clone())
        }
        rev
    }

    #[quickcheck]
    fn double_reversal_is_identity(xs: Vec<isize>) -> bool {
        xs == reverse(&reverse(&xs))
    }
}

fn setup_logger(level: String) -> Result<(), fern::InitError> {
    let colors = colors::ColoredLevelConfig::new()
        .info(colors::Color::Green)
        .warn(colors::Color::Magenta)
        .debug(colors::Color::Blue)
        .error(colors::Color::Red)
        .trace(colors::Color::BrightRed);
    fern::Dispatch::new()
        .format(move |out, message, record| {
            out.finish(format_args!(
                "[{}] {}",
                colors.color(record.level()),
                message
            ))
        })
        .level(LevelFilter::from_str(level.as_str()).unwrap())
        .chain(std::io::stdout())
        .apply()?;
    Ok(())
}

/// A basic example
#[derive(StructOpt, Debug)]
#[structopt(name = "MSL", about = "Simple interpreter to process MSL.")]
pub struct Opt {
    /// Logging level to show.
    ///  Possible values: [info, debug, warning, error, trace].
    #[structopt(short, long, name = "log-level", default_value = "info")]
    log_level: String,

    /// Allocated memory capacity.
    #[structopt(short, long, name = "allocated-memory", default_value = "100")]
    alloc: usize,

    /// File to execute.
    #[structopt(name = "FILE", parse(from_os_str))]
    file: Option<PathBuf>,
}

fn main() -> MSLResult<()> {
    let opt: Opt = Opt::from_args();
    setup_logger(opt.log_level).unwrap();
    let mut cpu = CPU::init(opt.alloc);
    if let Some(file_path) = opt.file {
        let res = interpreter::run_file(file_path, &mut cpu);
        if let Err(err) = res {
            error!("Error: {}", err.to_string());
        } else if let Ok(result) = res {
            info!("OK.");
            info!("{}", result);
        }
    } else {
        repl::start_repl_session()?;
    }
    Ok(())
    //    println!("{:#?}", opt);
}
